@@ -1,14 +1,13 @@
 """General module for parallelizing a dataframe apply function on a column (series) or entire row"""
 
 import logging
-from tokenize import group
-from tqdm import tqdm
-from multiprocessing import Pool
+from typing import Callable, Union, List
+from functools import partial
 import numpy as np
 import pandas as pd
+from tqdm import tqdm
+from pathos.multiprocessing import ProcessingPool as Pool
 from pandas.core.groupby.generic import DataFrameGroupBy
-from typing import Callable
-from functools import partial
 
 tqdm.pandas()
 
@@ -24,12 +23,6 @@ def parallelize_dataframe(df: pd.DataFrame, func: Callable, n_cores: int) -> pd.
     pool.join()
     return df
 
-def parallelize_groupby(df_grouped: DataFrameGroupBy, func: Callable, n_cores: int) -> pd.DataFrame:
-    with Pool(n_cores) as p:
-        ret_list = p.map(func, [group for _, group in df_grouped])
-    return pd.concat(ret_list)
-
-
 """
 Apply on series
 Usage:
@@ -123,22 +116,36 @@ Usage:
     serial => apply_on_groupby_parallel(df.groubpy([cols]), f, n_cores)
 """
 
-def _groupby_serial_func(data, func):
-    return [func(x) for x in tqdm(data)]
+def _groupby_serial_func(data: List, func: Callable, pbar: bool = True) -> List:
+    _range = tqdm(data) if pbar else data
+    return [func(x) for x in _range]
 
-def apply_on_groupby_parallel(df_grouped: DataFrameGroupBy, func: Callable, n_cores: int):
-    assert isinstance(df_grouped, DataFrameGroupBy)
+def apply_on_groupby_parallel(df_grouped: DataFrameGroupBy, func: Callable, n_cores: int, pbar: bool = True) \
+    -> Union[pd.DataFrame, pd.Series]:
+    X = iter(df_grouped)
     grouped_data = []
-
-    # groupby returns a tuple (key, sub_df), we only care about the sub_df. Other use-cases may appear later.
-    for item in iter(df_grouped):
+    key_data = []
+    for item in X:
+        key_data.append(item[0])
         grouped_data.append(item[1])
-
-    n_cores = min(len(grouped_data), n_cores)
+    # This is here to suppress warnings
+    grouped_data = np.array(grouped_data, dtype=object)
     split_data = np.array_split(grouped_data, n_cores)
-    pool = Pool(n_cores)
 
-    res = pool.map(partial(_groupby_serial_func, func=func), split_data)
-    res = np.concatenate(res, dtype=object)
-    res = pd.DataFrame(res)
+    # Run the multi-process job
+    pool = Pool(n_cores)
+    res = pool.map(partial(_groupby_serial_func, func=func, pbar=pbar), split_data)
+
+    # Concatenate the result to preserve the original result of a regular groupby pandas code.
+    concat_res = np.concatenate(res)
+
+    if len(concat_res[0].shape) > 0:
+        columns = res[0][0].index
+        res = pd.DataFrame(concat_res, columns=columns)
+    else:
+        res = pd.Series(concat_res)
+
+    # Fix the index
+    res.index = key_data
+    res.index.name = df_grouped.keys
     return res
